# -*- coding: utf-8 -*-

import json

from .exceptions import AppNotFoundError, ContainerNotFoundError, ImageNotFoundError
from .flock import lock_ex
from .config import REPO_FILE, LOCK_FILE

TYPE_APP = 'apps'
TYPE_CONTAINER = 'containers'
TYPE_IMAGE = 'images'

def load():
    try:
        with open(REPO_FILE) as f:
            return json.load(f)
    except FileNotFoundError:
        return {TYPE_IMAGE: {}, TYPE_CONTAINER: {}, TYPE_APP: {}}

def save(data):
    with open(REPO_FILE, 'w') as f:
        json.dump(data, f, sort_keys=True, indent=4)

def get_entries(entry_type):
    with lock_ex(LOCK_FILE):
        data = load()
    return data[entry_type]

def get_entry(entry_type, name):
    return get_entries(entry_type)[name]

def add_entry(entry_type, name, definition):
    with lock_ex(LOCK_FILE):
        data = load()
        data[entry_type][name] = definition
        save(data)

def delete_entry(entry_type, name):
    with lock_ex(LOCK_FILE):
        data = load()
        try:
            del data[entry_type][name]
            save(data)
        except KeyError:
            pass

def get_images():
    return get_entries(TYPE_IMAGE)

def get_image(image_name):
    try:
        return get_entry(TYPE_IMAGE, image_name)
    except KeyError as e:
        raise ImageNotFoundError(image_name) from e

def register_image(image_name, definition):
    add_entry(TYPE_IMAGE, image_name, definition)

def unregister_image(image_name):
    delete_entry(TYPE_IMAGE, image_name)

def get_containers():
    return get_entries(TYPE_CONTAINER)

def get_container(container_name):
    try:
        return get_entry(TYPE_CONTAINER, container_name)
    except KeyError as e:
        raise ContainerNotFoundError(container_name) from e

def register_container(container_name, definition):
    add_entry(TYPE_CONTAINER, container_name, definition)

def unregister_container(container_name):
    delete_entry(TYPE_CONTAINER, container_name)

def get_apps():
    return get_entries(TYPE_APP)

def get_app(app_name):
    try:
        return get_entry(TYPE_APP, app_name)
    except KeyError as e:
        raise ImageNotFoundError(app_name) from e

def register_app(app_name, definition):
    add_entry(TYPE_APP, app_name, definition)

def unregister_app(app_name):
    delete_entry(TYPE_APP, app_name)
