# -*- coding: utf-8 -*-

import json
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, utils
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from .exceptions import AppNotFoundError, ImageNotFoundError
from .flock import lock_ex
from .config import PUB_LOCK_FILE, PUB_PRIVKEY_FILE, PUB_REPO_FILE, PUB_SIG_FILE

TYPE_APP = 'apps'
TYPE_IMAGE = 'images'

class TarSizeCounter:
    def __init__(self):
        self.size = 0

    def add_file(self, tarinfo):
        self.size += tarinfo.size
        return tarinfo

def sign_file(file_path):
    # Generate ECDSA HMAC SHA512 signature of a file using EC private key
    sha512 = hashes.SHA512()
    hasher = hashes.Hash(sha512, default_backend())
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(64*1024)
            if not data:
                break
            hasher.update(data)
    with open(PUB_PRIVKEY_FILE, 'rb') as f:
        private_key = load_pem_private_key(f.read(), None, default_backend())
    return private_key.sign(hasher.finalize(), ec.ECDSA(utils.Prehashed(sha512)))

def load():
    try:
        with open(PUB_REPO_FILE) as f:
            return json.load(f)
    except FileNotFoundError:
        return {TYPE_IMAGE: {}, TYPE_APP: {}}

def save(data):
    with open(PUB_REPO_FILE, 'w') as f:
        json.dump(data, f, sort_keys=True, indent=4)
    # Cryptographically sign the repository file
    signature = sign_file(PUB_REPO_FILE)
    with open(PUB_SIG_FILE, 'wb') as f:
        f.write(signature)

def get_entries(entry_type):
    with lock_ex(PUB_LOCK_FILE):
        data = load()
    return data[entry_type]

def get_entry(entry_type, name):
    return get_entries(entry_type)[name]

def add_entry(entry_type, name, definition):
    with lock_ex(PUB_LOCK_FILE):
        data = load()
        data[entry_type][name] = definition
        save(data)

def delete_entry(entry_type, name):
    with lock_ex(PUB_LOCK_FILE):
        data = load()
        try:
            del data[entry_type][name]
            save(data)
        except KeyError:
            pass

def get_images():
    return get_entries(TYPE_IMAGE)

def get_image(image_name):
    try:
        return get_entry(TYPE_IMAGE, image_name)
    except KeyError as e:
        raise ImageNotFoundError(image_name) from e

def register_image(image_name, definition):
    add_entry(TYPE_IMAGE, image_name, definition)

def unregister_image(image_name):
    delete_entry(TYPE_IMAGE, image_name)

def get_apps():
    return get_entries(TYPE_APP)

def get_app(app_name):
    try:
        return get_entry(TYPE_APP, app_name)
    except KeyError as e:
        raise ImageNotFoundError(app_name) from e

def register_app(app_name, definition):
    add_entry(TYPE_APP, app_name, definition)

def unregister_app(app_name):
    delete_entry(TYPE_APP, app_name)
